#include "abup_custom_timer.h"
#include <string.h>
#include "hal/clk_gate_ll.h"
#include "hal/interrupt_controller_hal.h"
#include "hal/cpu_hal.h"


typedef struct {
    timer_hal_context_t hal;
} timer_obj_t;

static const char *ABUP_TIMER_TAG = "abup_timer_group";
static timer_obj_t abup_timer_obj[TIMER_GROUP_MAX][TIMER_MAX] = {0};

//This bitmask has an 1 if the int should be disabled when the flash is disabled.
//static uint32_t non_iram_int_mask[SOC_CPU_CORES_NUM];

volatile int abup_timer_callback_cnt = 0;
static void abup_timer_group_isr_callback(void *args)
{
    //ESP_LOGI(ABUP_TIMER_TAG, "=== abup_timer_group_isr_callback calling ===");
    abup_timer_callback_cnt++;
#if 0
    BaseType_t high_task_awoken = pdFALSE;
    abup_timer_info_t *info = (abup_timer_info_t *) args;

    uint64_t timer_counter_value = timer_group_get_counter_value_in_isr(info->timer_group, info->timer_idx);

    /* Prepare basic event data that will be then sent back to task */
    example_timer_event_t evt = {
        .info.timer_group = info->timer_group,
        .info.timer_idx = info->timer_idx,
        .info.auto_reload = info->auto_reload,
        .info.alarm_interval = info->alarm_interval,
        .timer_counter_value = timer_counter_value
    };

    if (!info->auto_reload) {
        timer_counter_value += info->alarm_interval * TIMER_SCALE;
        timer_group_set_alarm_value_in_isr(info->timer_group, info->timer_idx, timer_counter_value);
    }

    /* Now just send the event data back to the main program task */
    xQueueSendFromISR(s_timer_queue, &evt, &high_task_awoken);
#endif
    // re-start timer
    abup_timer_group_clr_intr_status_in_isr(TIMER_GROUP_0, TIMER_0);
    abup_timer_group_enable_alarm_in_isr(TIMER_GROUP_0, TIMER_0);
}

void abup_tg_timer_init(int group, int timer, bool auto_reload, int timer_interval_msec)
{
    /* Select and initialize basic parameters of the timer */
    timer_config_t config = {
        .divider = TIMER_DIVIDER,
        .counter_dir = TIMER_COUNT_UP,
        .counter_en = TIMER_PAUSE,
        .alarm_en = TIMER_ALARM_EN,
        .auto_reload = auto_reload,
    }; // default clock source is APB
    ESP_LOGI(ABUP_TIMER_TAG, "abup_tg_timer_init is running");
    abup_timer_init(group, timer, &config);

    /* Timer's counter will initially start from value below.
       Also, if auto_reload is set, this value will be automatically reload on alarm */
    abup_timer_set_counter_value(group, timer, 0);

    /* Configure the alarm value and the interrupt on alarm. */
    abup_timer_set_alarm_value(group, timer, VAL_MS(timer_interval_msec));
    abup_timer_enable_intr(group, timer);
    ESP_LOGI(ABUP_TIMER_TAG, "abup_timer_isr_callback_add is running");
    abup_timer_isr_callback_add(group, timer, abup_timer_group_isr_callback, NULL);

    abup_timer_start(group, timer);
}


esp_err_t abup_timer_init(timer_group_t group_num, timer_idx_t timer_num, const timer_config_t *config)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(config != NULL, TIMER_PARAM_ADDR_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(config->divider > 1 && config->divider < 65537, DIVIDER_RANGE_ERROR, ESP_ERR_INVALID_ARG);

    periph_ll_enable_clk_clear_rst(timer_group_periph_signals.groups[group_num].module);

    timer_hal_init(&(abup_timer_obj[group_num][timer_num].hal), group_num, timer_num);
    timer_hal_intr_disable(&(abup_timer_obj[group_num][timer_num].hal));
    timer_hal_clear_intr_status(&(abup_timer_obj[group_num][timer_num].hal));
    timer_hal_set_auto_reload(&(abup_timer_obj[group_num][timer_num].hal), config->auto_reload);
    timer_hal_set_divider(&(abup_timer_obj[group_num][timer_num].hal), config->divider);
    timer_hal_set_counter_increase(&(abup_timer_obj[group_num][timer_num].hal), config->counter_dir);
    timer_hal_set_alarm_enable(&(abup_timer_obj[group_num][timer_num].hal), config->alarm_en);
    timer_hal_set_level_int_enable(&(abup_timer_obj[group_num][timer_num].hal), true);
    if (config->intr_type != TIMER_INTR_LEVEL) {
        ESP_LOGI(ABUP_TIMER_TAG, "only support Level Interrupt, switch to Level Interrupt instead");
    }
    timer_hal_set_counter_enable(&(abup_timer_obj[group_num][timer_num].hal), config->counter_en);
#if SOC_TIMER_GROUP_SUPPORT_XTAL
    timer_hal_set_use_xtal(&(abup_timer_obj[group_num][timer_num].hal), config->clk_src);
#endif

    return ESP_OK;
}

esp_err_t abup_timer_deinit(timer_group_t group_num, timer_idx_t timer_num)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_set_counter_enable(&(abup_timer_obj[group_num][timer_num].hal), TIMER_PAUSE);
    timer_hal_intr_disable(&(abup_timer_obj[group_num][timer_num].hal));
    timer_hal_clear_intr_status(&(abup_timer_obj[group_num][timer_num].hal));

    memset(&abup_timer_obj[group_num][timer_num], 0, sizeof(timer_obj_t));

    return ESP_OK;
}

esp_err_t abup_timer_set_counter_value(timer_group_t group_num, timer_idx_t timer_num, uint64_t load_val)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_set_counter_value(&(abup_timer_obj[group_num][timer_num].hal), load_val);

    return ESP_OK;
}

esp_err_t abup_timer_set_alarm_value(timer_group_t group_num, timer_idx_t timer_num, uint64_t alarm_value)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_set_alarm_value(&(abup_timer_obj[group_num][timer_num].hal), alarm_value);

    return ESP_OK;
}

esp_err_t abup_timer_enable_intr(timer_group_t group_num, timer_idx_t timer_num)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_intr_enable(&(abup_timer_obj[group_num][timer_num].hal));

    return ESP_OK;
}

esp_err_t abup_timer_disable_intr(timer_group_t group_num, timer_idx_t timer_num)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_intr_disable(&(abup_timer_obj[group_num][timer_num].hal));

    return ESP_OK;
}


esp_err_t abup_timer_start(timer_group_t group_num, timer_idx_t timer_num)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_set_counter_enable(&(abup_timer_obj[group_num][timer_num].hal), TIMER_START);

    return ESP_OK;
}

esp_err_t abup_timer_pause(timer_group_t group_num, timer_idx_t timer_num)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    timer_hal_set_counter_enable(&(abup_timer_obj[group_num][timer_num].hal), TIMER_PAUSE);

    return ESP_OK;
}

esp_err_t abup_timer_isr_callback_add(timer_group_t group_num, timer_idx_t timer_num, intr_handler_t isr_handler, void *args)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    abup_timer_disable_intr(group_num, timer_num);

    if(abup_timer_isr_register(group_num, timer_num, isr_handler, args) != ESP_OK)
    {
        ESP_LOGW("ABUP", "abup_timer_isr_register ERROR");
        return ESP_FAIL;
    }

    abup_timer_enable_intr(group_num, timer_num);

    return ESP_OK;
}


esp_err_t abup_timer_isr_register(timer_group_t group_num, timer_idx_t timer_num, void (*fn)(void *), void *arg)
{
    TIMER_CHECK(group_num < TIMER_GROUP_MAX, TIMER_GROUP_NUM_ERROR, ESP_ERR_INVALID_ARG);
    TIMER_CHECK(timer_num < TIMER_MAX, TIMER_NUM_ERROR, ESP_ERR_INVALID_ARG);

    uint32_t status_reg = 0;
    uint32_t mask = 0;
    timer_hal_get_status_reg_mask_bit(&(abup_timer_obj[group_num][timer_num].hal), &status_reg, &mask);
    
    return abup_esp_intr_alloc_intrstatus(timer_group_periph_signals.groups[group_num].t0_irq_id + timer_num, status_reg, mask, fn, arg);
}

esp_err_t abup_esp_intr_alloc_intrstatus(int source, uint32_t intrstatusreg, uint32_t intrstatusmask, intr_handler_t handler, void *arg)
{
    int force=-1;
    int intr_num = 11;//固定使用中断号11
    ESP_LOGI(ABUP_TIMER_TAG, "esp_intr_alloc_intrstatus (cpu %d): checking args", cpu_hal_get_core_id());

    //Statusreg should have a mask
    if (intrstatusreg && !intrstatusmask) return ESP_ERR_INVALID_ARG;

    int cpu=cpu_hal_get_core_id();
    
    if (source>=0) {
        intr_matrix_set(cpu, source, intr_num);
    }
    
    if (handler) {
        interrupt_controller_hal_set_int_handler(intr_num, handler, arg);
    }

    //non_iram_int_mask[cpu]|=(1<<intr_num);

    //Enable int at CPU-level;
    ABUP_ESP_INTR_ENABLE(intr_num);

#ifdef SOC_CPU_HAS_FLEXIBLE_INTC
    //Extract the level from the interrupt passed flags
    int level = __builtin_ffs(0) + 1;
    interrupt_controller_hal_set_int_level(intr_num, level);
    interrupt_controller_hal_set_int_type(intr_num, INTTP_LEVEL);
#endif

    ESP_LOGI(ABUP_TIMER_TAG, "Connected src %d to int %d (cpu %d)", source, intr_num, cpu);
    return ESP_OK;
}


//These functions are provided in ROM, but the ROM-based functions use non-multicore-capable
//virtualized interrupt levels. Thus, we disable them in the ld file and provide working
//equivalents here.


void IRAM_ATTR abup_ets_isr_unmask(uint32_t mask) {
    interrupt_controller_hal_enable_interrupts(mask);
}

void IRAM_ATTR abup_ets_isr_mask(uint32_t mask) {
    interrupt_controller_hal_disable_interrupts(mask);
}

void abup_esp_intr_enable_source(int inum)
{
    interrupt_controller_hal_enable_interrupts(1 << inum);
}

void abup_esp_intr_disable_source(int inum)
{
    interrupt_controller_hal_disable_interrupts(1 << inum);
}

void IRAM_ATTR abup_timer_group_clr_intr_status_in_isr(timer_group_t group_num, timer_idx_t timer_num)
{
    timer_hal_clear_intr_status(&(abup_timer_obj[group_num][timer_num].hal));
}

void IRAM_ATTR abup_timer_group_enable_alarm_in_isr(timer_group_t group_num, timer_idx_t timer_num)
{
    timer_hal_set_alarm_enable(&(abup_timer_obj[group_num][timer_num].hal), true);
}


