#ifndef __ABUP_CUSTOM_TIMER_H
#define __ABUP_CUSTOM_TIMER_H

#include "esp_attr.h"
#include "esp_err.h"
#include "esp_log.h"
#include "soc/soc.h"
#include "soc/timer_periph.h"
#include "hal/timer_hal.h"
#include "hal/timer_types.h"
#include "riscv/interrupt.h"


    
#define TIMER_GROUP_NUM_ERROR   "TIMER GROUP NUM ERROR"
#define TIMER_NUM_ERROR         "HW TIMER NUM ERROR"
#define TIMER_PARAM_ADDR_ERROR  "HW TIMER PARAM ADDR ERROR"
#define TIMER_NEVER_INIT_ERROR  "HW TIMER NEVER INIT ERROR"
#define TIMER_COUNT_DIR_ERROR   "HW TIMER COUNTER DIR ERROR"
#define TIMER_AUTORELOAD_ERROR  "HW TIMER AUTORELOAD ERROR"
#define TIMER_SCALE_ERROR       "HW TIMER SCALE ERROR"
#define TIMER_ALARM_ERROR       "HW TIMER ALARM ERROR"
#define DIVIDER_RANGE_ERROR     "HW TIMER divider outside of [2, 65536] range error"


#define TIMER_CHECK(a, str, ret_val) \
    if (!(a)) { \
        ESP_LOGE(ABUP_TIMER_TAG,"%s(%d): %s", __FUNCTION__, __LINE__, str); \
        return (ret_val); \
    }

#define TIMER_BASE_CLK   (CPU_CLK_FREQ)  /*!< Frequency of the clock on the input of the timer groups */
#define TIMER_DIVIDER         (16)  //  Hardware timer clock divider
#define SEC_TO_MICRO_SEC(x)    ((x) / 1000)                  // Convert second to millisecond
#define TIMER_TICKS            (TIMER_BASE_CLK / TIMER_DIVIDER)     // TIMER_BASE_CLK = APB_CLK = 80MHz
#define VAL_MS(x)           SEC_TO_MICRO_SEC(x * TIMER_TICKS)     // Alarm value in micro-seconds

/** Enable interrupt by interrupt number */
#define ABUP_ESP_INTR_ENABLE(inum)  abup_esp_intr_enable_source(inum)

/** Disable interrupt by interrupt number */
#define ABUP_ESP_INTR_DISABLE(inum) abup_esp_intr_disable_source(inum)

extern volatile int abup_timer_callback_cnt;

void abup_tg_timer_init(int group, int timer, bool auto_reload, int timer_interval_sec);

esp_err_t abup_timer_init(timer_group_t group_num, timer_idx_t timer_num, const timer_config_t *config);
esp_err_t abup_timer_deinit(timer_group_t group_num, timer_idx_t timer_num);
esp_err_t abup_timer_set_counter_value(timer_group_t group_num, timer_idx_t timer_num, uint64_t load_val);
esp_err_t abup_timer_set_alarm_value(timer_group_t group_num, timer_idx_t timer_num, uint64_t alarm_value);
esp_err_t abup_timer_enable_intr(timer_group_t group_num, timer_idx_t timer_num);
esp_err_t abup_timer_disable_intr(timer_group_t group_num, timer_idx_t timer_num);
esp_err_t abup_timer_start(timer_group_t group_num, timer_idx_t timer_num);
esp_err_t abup_timer_pause(timer_group_t group_num, timer_idx_t timer_num);
esp_err_t abup_timer_isr_callback_add(timer_group_t group_num, timer_idx_t timer_num, intr_handler_t isr_handler, void *args);
esp_err_t abup_timer_isr_register(timer_group_t group_num, timer_idx_t timer_num, void (*fn)(void *), void *arg);
esp_err_t abup_esp_intr_alloc_intrstatus(int source, uint32_t intrstatusreg, uint32_t intrstatusmask, intr_handler_t handler, void *arg);

/**
 * @brief enable the interrupt source based on its number
 * @param inum interrupt number from 0 to 31
 */
void abup_esp_intr_enable_source(int inum);

/**
 * @brief disable the interrupt source based on its number
 * @param inum interrupt number from 0 to 31
 */
void abup_esp_intr_disable_source(int inum);
void IRAM_ATTR abup_timer_group_clr_intr_status_in_isr(timer_group_t group_num, timer_idx_t timer_num);
void IRAM_ATTR abup_timer_group_enable_alarm_in_isr(timer_group_t group_num, timer_idx_t timer_num);

#endif

